//HVTC-ted converter by Hermit (Mihaly Horvath) (Year 2023)

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <ctype.h>
#include <libgen.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#define DEBUG 1
#define HTMLFOLDER "html"
#define INFOFOLDER "html/info"
#define MISSINGINFO_FOLDER "edits/missing-info"
#define EDITFOLDER "edits"
#define TARGETFOLDER "HVTC-ted"
#define INFOTARGETFOLDER "HVTC-info"
#define TARGETEXTENSION ".ted" //".ted.prg"
#define INFOEXTENSION ".txt"
#define BASICSTART_ADDRESS 0x1001
#define DEFAULT_INITADDRESS 4109 //$100D //mostly used, used as fallback
#define KEYMATRIX_SIZE 64
#define NOKEY 0xFF
#define FIELDSTRING_MAX_SIZE 1000
#define INFOFILE_MAX_SIZE 10000 //2000
#define MUSICFILE_MAX_SIZE 65536
#define HTMLFILE_MAX_SIZE 100000
#define PATH_MAX_LENGTH 1024
#define FRAMESPEED_MAX 8
#define DEFAULT_KEYDELAY 10 //*100ms

enum cRTED_Specifications { TED_INFOFIELD_SIZE=32, SUBTUNE_AMOUNT_MAX=255 };

typedef struct cRTED_TEDheader {                   //Offset:      default/info: //TMF-format header within Hermit's custom TED-format container
 unsigned char LoadAddressL,LoadAddressH;         //0,1        PRG load-address, always $1001 to be runnable from BASIC prompt (Spetro's tunes are at $1000 though)
 unsigned char BASICnextLineL,BASICnextLineH;     //2,3        next BASIC line's address ($100B) (but for e.g. Csabo's paratroopers it's not correct), can't count on it
 unsigned char BASIClineNumberL,BASIClineNumberH; //4,5        BASIC SYS-command's line-number, POINTER TO THE TED(TMF)-HEADER
 unsigned char BASICsysCommand;                   //6          BASIC SYS-command token ($9E)
 unsigned char BASICsysParameter[4];              //7..10      BASIC SYS-parameter, the address of runnable machine code in ASCII
 unsigned char BASIClineEnd;                      //11     $0B BASIC line-end ($00)
 unsigned char BASICend[2];                       //12,13  $0C BASIC program end ($00,$00)
 unsigned char Filler[3];                         //14..16 $0E filler empty places (0,0,0)
 //TMF-only part
 char MagicString[8];                             //17..24 $11 "TEDMUSIC" signature/MagicString
 unsigned char Version;                           //25     $19 Preliminary specification-version (0)
 unsigned char DataOffsetL,DataOffsetH;           //26,27  $1A Where the music (maybe with loadaddress) resides - relative to beginning of file
 unsigned char DataAddressL,DataAddressH;         //28,29  $1C Where the music should be loaded/copied
 unsigned char InitAddressL,InitAddressH;         //30,31  $1E Init-routine to start/set song, subtune in Accu. Init mustn't change IRQ
 unsigned char PlayAddressL,PlayAddressH;         //32,33  $20 Play-routine's address, don't disturb IRQ (0=RTED?)
 unsigned char SubtuneAmount;                     //34     $22 amount of subtunes ('1' if only 1)
 unsigned char TimingSource;                      //35     $23 0:PAL Vblank(50Hz), 1:NTSC Vblank(60Hz), 2:PAL Timer1, 3:NTSC Timer1, 4:PAL Timer1(set by init), 5:NTSC Timer1(set by init)
 unsigned char Timer1valueL,Timer1valueH;         //36,37  $24 TED timer1 value (PAL: (1x:17734), 2x:8867, 4x:4433) (NTSC: (1x:14915), 2x:7457, 4x:3728)
 unsigned char SongFlags;                         //38     $26 bit4:FM?, bit3:AY?, bit2:DigiBlaster?, bit1:TEDcard?, bit0:Screen-off?
 unsigned char FileFlags;                         //39     $27 bit0:SongLength (duration) block is present (2 byte-words for all subtunes: seconds), bit1:keypress-block is present
 unsigned char DurationOffsetL,DurationOffsetH;   //40,41  $28 file-offset of the duration(songlength)-block (and possible keypress-block following it)
 unsigned char Reserved[23];                      //42..64 $2A reserved for future use (padding to have coming textfields aligned to $10 in memory)
 unsigned char Title[TED_INFOFIELD_SIZE];         //65..96 $41 Title of music
 unsigned char Author[TED_INFOFIELD_SIZE];        //97...  $61 Author/composer of music
 unsigned char ReleaseInfo[TED_INFOFIELD_SIZE];   //129..  $81 Release-info, recommended: [Year] [Group]
 unsigned char Tool[TED_INFOFIELD_SIZE];          //161..  $A1 Tracker/composer-tool that the music was created with
 unsigned char ReservedInfoField[64];             //193..  $C1 Reserved textfield for possible future use
 //(257..=$101.. built-in starter/initializer routine for pure TMF,) duration-block, etc.
} cRTED_TEDheader;


unsigned char *KeyMatrixSymbols[KEYMATRIX_SIZE] = {
//  0       1         2       3       4      5     6     7     8       9       A       B     C        D     E     F  //$FF08 / $FD30 bits
 "Del",  "Return", "Pound", "HELP",  "F1",  "F2", "F3", "@",  "3",    "W",    "A",    "4",  "Z",     "S",  "E", "Shift",  //0x //$FE,$FD
 "5",    "R",      "D",     "6",     "C",   "F",  "T",  "X",  "7",    "Y",    "G",    "8",  "B",     "H",  "U", "V",      //1x //$FB,$F7
 "9",    "I",      "J",     "0",     "M",   "K",  "O",  "N",  "Down", "P",    "L",    "Up", ".",     ":",  "-", ",",      //2x //$EF,$DF
 "Left", "*",      ";",     "Right", "Esc", "=",  "+",  "/",  "1",    "Home", "Ctrl", "2",  "Space", "C=", "Q", "STOP" }; //3x //$BF,$7F

                                                        //      1x   2x   3x   4x    5x   6x   7x   8x
const unsigned char FrameSpeedTimerHi[1+FRAMESPEED_MAX] = { 0, 0x45,0x22,0x17,0x11, 0x0d,0x0b,0x09,0x08 };
const unsigned char FrameSpeedTimerLo[1+FRAMESPEED_MAX] = { 0, 0x46,0xA3,0x17,0x51, 0xda,0x8b,0xe5,0xA8 };


char compareFileExtension(char *filename, char *extension) { //get pointer of file-extension from filename string  //if no '.' found, point to end of the string
 short i,j;
 for (i=strlen(filename)-1,j=strlen(extension)-1; i>=0 && j>=0; --i,--j) if(tolower(filename[i])!=tolower(extension[j])) return 0;
 return 1;
}
char* filExt(char *filename) { //get pointer of file-extension from filename string  //if no '.' found, point to end of the string
 char* LastDotPos = strrchr(filename,'.'); if (LastDotPos == NULL) return (filename+strlen(filename)); //make strcmp not to find match, otherwise it would be segmentation fault
 return LastDotPos;
}
void cutExt(char *filename) { *filExt(filename)='\0'; } //cut the extension of the filename by putting 0 at position of '.'  //omit original extension with 0 string-delimiter
void changeExt (char *filename, char* extension) {
 cutExt(filename); strcat(filename,extension);
}


void processHTMLinfo() {
 char TextTemp; DIR *Dir; struct dirent *Entry;
 FILE *HTMLfile, *EditFile, *OutFile, *OutFile2;
 static char DataIn, Path[PATH_MAX_LENGTH], HTMLdata[HTMLFILE_MAX_SIZE], EditText[INFOFILE_MAX_SIZE]; static char *EmptyString="";
 int i, HTMLfileSize, EditFileSize; char *TitleText, *TitleTextEnd, *InfoText, *TextPtr, *OutFileName;

 char* parseHTMLvalue (char* fieldname) {
  int x; unsigned char *FoundText, *FoundTextEnd;
  static char FieldName[FIELDSTRING_MAX_SIZE]="", Value[FIELDSTRING_MAX_SIZE]="";
  strcpy(FieldName,fieldname); strcat(FieldName,"</td><td");
  FoundText = strstr(InfoText,FieldName); if (FoundText==NULL) return EmptyString;
  FoundText+=strlen(FieldName); FoundTextEnd = strstr(FoundText,"</"); if (FoundTextEnd==NULL) return EmptyString;
  for (x=0 ;FoundText+x < FoundTextEnd; ++x) Value[x] = FoundText[x];
  Value[x] = '\0'; return strrchr(Value,'>')+1;
 }

 void findSaveValue (char* htmlfieldname, char* infofieldname) {
  char *ValuePtr, *ValuePtr2; static char FieldName[FIELDSTRING_MAX_SIZE];
  ValuePtr = parseHTMLvalue(htmlfieldname);
  strcpy(FieldName,infofieldname); strcat(FieldName,": "); ValuePtr2 = strstr(EditText,FieldName);
  if (ValuePtr2==NULL || ValuePtr2[strlen(FieldName)]=='\n' || !ValuePtr2[strlen(FieldName)]) ValuePtr2 = EmptyString;
  if (ValuePtr!=EmptyString) fprintf(OutFile, "%s: %s\n", infofieldname, ValuePtr);
  else if (ValuePtr2==EmptyString) fprintf(OutFile2, "%s: %s\n", infofieldname, ValuePtr);
 }

 Dir = opendir(HTMLFOLDER); if (!Dir) { printf("Couldn't find html folder" HTMLFOLDER); exit(-1); }
 while ( (Entry=readdir(Dir)) != NULL ) {
  if ( compareFileExtension(Entry->d_name,".html") && strcmp(Entry->d_name,".") && strcmp(Entry->d_name,"..") ) {
   strcpy(Path,HTMLFOLDER); strcat(Path,"/"); strcat(Path,Entry->d_name); if(DEBUG) printf("%s\n",Path);
   HTMLfile = fopen(Path,"r"); HTMLfileSize=0; while ((DataIn=fgetc(HTMLfile))!=EOF) HTMLdata[HTMLfileSize++]=DataIn;  HTMLdata[HTMLfileSize]='\0'; fclose(HTMLfile);
   InfoText = strstr(HTMLdata,"<!-- REQUEST_URI /hvtc/");
   TitleText = strstr(HTMLdata, "<title>HVTC - ")+14; TitleTextEnd = strstr(TitleText," - Plus/4 World</title>"); *TitleTextEnd='\0';

   if(DEBUG) printf("Title: %s\n",TitleText); OutFileName = parseHTMLvalue("Link");
   if(DEBUG) printf("Link: %s\n", OutFileName); changeExt(OutFileName,".txt");
   strcpy(Path,"mkdir -p "); strcat(Path,INFOFOLDER); strcat(Path,"/"); strcat(Path,OutFileName); dirname(Path); system(Path);
   strcpy(Path,INFOFOLDER); strcat(Path,"/"); strcat(Path,OutFileName);
   OutFile = fopen(Path,"w"); if(OutFile==NULL) { printf("Couldn't create file: %s\n",Path); continue; }
   strcpy(Path,"mkdir -p "); strcat(Path,MISSINGINFO_FOLDER); strcat(Path,"/"); strcat(Path,OutFileName); dirname(Path); system(Path);
   strcpy(Path,MISSINGINFO_FOLDER); strcat(Path,"/"); strcat(Path,OutFileName);
   OutFile2 = fopen(Path,"w"); if(OutFile2==NULL) { printf("Couldn't create file: %s\n",Path); continue; }
   strcpy(Path,EDITFOLDER); strcat(Path,"/"); strcat(Path,OutFileName);
   EditFile = fopen(Path,"r"); EditFileSize=0; EditText[0]='\0';
   if(EditFile!=NULL) { while ((DataIn=fgetc(EditFile))!=EOF) EditText[EditFileSize++]=DataIn;  EditText[EditFileSize]='\0'; fclose(EditFile); }

   fprintf(OutFile,"InfoSource: %s\n", Entry->d_name);
   fprintf(OutFile,"Title: %s\n", TitleText); findSaveValue("Music by","Author");
   findSaveValue("Release Date","Date"); findSaveValue("Player","Tool");
   sprintf(Path,"Song 2</td>"); TextPtr=strstr(InfoText,Path);  if (TextPtr!=NULL) { TextTemp = *TextPtr; *TextPtr='\0'; } //info only until next subtune
   findSaveValue("Cover of","Covers"); findSaveValue("Notes","Notes");
   findSaveValue("Channels","Channels"); findSaveValue("Speed","FrameSpeed");
   findSaveValue("Length","PlayTime"); findSaveValue("Loops","Looping");
   if (TextPtr!=NULL) *TextPtr = TextTemp;
   for (i=2; i<=SUBTUNE_AMOUNT_MAX; ++i) {
    sprintf(Path,"Song %d</td>",i); InfoText = strstr(InfoText,Path);
    if (InfoText==NULL) break;
    else {
     sprintf(Path,"Song %d</td>",i+1); TextPtr=strstr(InfoText,Path);
     if (TextPtr!=NULL) { TextTemp = *TextPtr; *TextPtr='\0'; } //info only until next subtune
     sprintf(Path,"Covers%d",i); findSaveValue("Cover of",Path); //sprintf(Path,"Notes%d",i); findSaveValue("Notes",Path);
     sprintf(Path,"Channels%d",i); findSaveValue("Channels",Path);
     sprintf(Path,"FrameSpeed%d",i); findSaveValue("Speed",Path);
     sprintf(Path,"PlayTime%d",i); findSaveValue("Length",Path);
     sprintf(Path,"Looping%d",i); findSaveValue("Loops",Path);
     if (TextPtr!=NULL) *TextPtr = TextTemp;
   }}
   if (i-1>1) fprintf(OutFile,"SubTunes: %d\n",i-1);
   fprintf(OutFile,"\n"); fclose(OutFile);
 }} printf("\n");
 closedir(Dir);
}


void convertToTEDfile (char* path) {
 FILE *InFile, *InfoFile, *EditFile, *OutFile, *OutFile2;
 int i,m, Tmp, Minutes, Seconds, DataIn, LoadAddress, BASICoffset=0, FileSize=0, InfoFileSize=0, EditFileSize=0; char* InfoPtr;
 static unsigned char FileData[MUSICFILE_MAX_SIZE], InfoText[INFOFILE_MAX_SIZE], EditText[INFOFILE_MAX_SIZE], HVTCinfo[INFOFILE_MAX_SIZE]; static cRTED_TEDheader TEDheader; static char *EmptyString="";
 static int TEDfileCount=0;
 static char TargetPath[PATH_MAX_LENGTH]="", TargetPath2[PATH_MAX_LENGTH]="", InfoPath[PATH_MAX_LENGTH]="", TmpString[FIELDSTRING_MAX_SIZE]="";

 char* findInfo (char* fieldname, char format) { //format text (if target is TEDfile-field) or not?
  int l,x; unsigned char chrlen; unsigned char* ValuePtr; static char FieldName[FIELDSTRING_MAX_SIZE], Value[FIELDSTRING_MAX_SIZE];

  unsigned char UTF8toASCII (unsigned char chr) {
   unsigned char trial; int j,k, codepoint;
   static const int ISO8859_2_codepoints[]={ //starting from 0xA0, the rest is ASCII
    0x0A0,0x104,0x2d8,0x141,0x0A4,0x13D,0x15A,0x0A7,0x0A8,0x160,0x15E,0x164,0x179,0x0AD,0x17D,0x17B,
    0x0B0,0x105,0x2DB,0x142,0x0B4,0x13E,0x15B,0x2C7,0x0B8,0x161,0x15F,0x165,0x17A,0x2DD,0x17E,0x17C,
    0x154,0x0C1,0x0C2,0x102,0x0C4,0x139,0x106,0x0C7,0x10C,0x0C9,0x118,0x0CB,0x11A,0x0CD,0x0CE,0x10E,
    0x110,0x143,0x147,0x0D3,0x0D4,0x150,0x0D6,0x0D7,0x158,0x16E,0x0DA,0x170,0x0DC,0x0DD,0x162,0x0DF,
    0x155,0x0E1,0x0E2,0x103,0x0E4,0x13A,0x107,0x0E7,0x10D,0x0E9,0x119,0x0EB,0x11B,0x0ED,0x0EE,0x10F,
    0x111,0x144,0x148,0x0F3,0x0F4,0x151,0x0F6,0x0F7,0x159,0x16F,0x0FA,0x171,0x0FC,0x0FD,0x163,0x2D9 };
   chrlen=1;
   if (chr>=0x80) {
    if (0xC0<=chr && chr<0xE0) chrlen=2; //2-byte long UTF8 char
    else if (0xE0<=chr && chr<0xF0) chrlen=3; //3-byte long
    else if (0xF0<=chr && chr<0xF8) chrlen=4; //4-byte long
    if(chrlen>1) { //get codepoint of valid UTF-8 character
     codepoint = chr&0x3F;
     for (j=1; j<chrlen; j++) codepoint = codepoint*0x40 + (ValuePtr[x+j] & 0x3F);
     if(codepoint<0x100) chr=codepoint;
     else { //detect if codepaged character is included in the array of characters below 0x100
      for(j=0;j<0x60;j++) { if (codepoint==ISO8859_2_codepoints[j]) {chr=j+0xA0; break;} }
   }}}
   return chr;
  }

  strcpy(FieldName,fieldname); strcat(FieldName,": ");
  ValuePtr = strstr(EditText,FieldName);
  if (ValuePtr==NULL || ValuePtr[strlen(FieldName)]=='\n' || !ValuePtr[strlen(FieldName)]) ValuePtr = strstr(InfoText,FieldName);
  if (ValuePtr!=NULL) {
   ValuePtr += strlen(FieldName);
   if (format) {
    for (l=x=0; l<TED_INFOFIELD_SIZE && ValuePtr[x]!='\n' && ValuePtr[x]; ++l,++x) { Value[l] = UTF8toASCII(ValuePtr[x]); x += chrlen-1; }
   }
   else { for (l=0; ValuePtr[l]!='\n' && ValuePtr[l]; ++l) Value[l]=ValuePtr[l]; }
   Value[l]='\0'; return Value;
  }
  else return EmptyString;
 }

 void fillInfoText (char* info, char* fieldname, char* target) { //fill in info-field in TMF-header with the given text
  char* ValuePtr;
  ValuePtr = findInfo(fieldname,1); if (ValuePtr!=EmptyString) info = ValuePtr; //use priorized alternative found in 'info'-folder txt files
  char PrevChr=' ';
  for (i=0; i<TED_INFOFIELD_SIZE && info[i]!='\0'; ++i) {
   target[i] = (PrevChr==' ' && ValuePtr==EmptyString)? toupper(info[i]) : info[i];
   if (target[i]=='_') target[i]=' '; PrevChr=target[i];
  }
  for (; i<TED_INFOFIELD_SIZE; ++i) target[i]=' ';
 }

 void saveInfoText (char* fieldname) {
  InfoPtr = findInfo(fieldname,0); if (InfoPtr!=EmptyString) { strcat(HVTCinfo,fieldname); strcat(HVTCinfo,": "); strcat(HVTCinfo,InfoPtr); strcat(HVTCinfo,"\n"); } //fprintf(OutFile2,"%s: %s\n",fieldname,InfoPtr);
 }

 void parseSavePlayTime () {
  if (InfoPtr==EmptyString) { Minutes=Seconds=0; }
  else if (!strcmp(InfoPtr,"Unlimited")) { Minutes=59; Seconds=59; }
  else { Tmp = sscanf(InfoPtr,"%d:%d",&Minutes,&Seconds); if (Tmp!=2) Minutes=Seconds=0; }
  Tmp = Minutes*60 + Seconds; fputc(Tmp&0xFF,OutFile); fputc(Tmp>>8,OutFile);
 }

 char caseInsensitiveStringCompare (char* string1, char* string2) {
  int i;
  for (i=0; i<FIELDSTRING_MAX_SIZE; ++i) {
   if (tolower(string1[i]) != tolower(string2[i])) return 0; //if differs or one is shorter
   if (string1[i]=='\0' && string2[i]=='\0') break;
  }
  return 1;
 }

 if(DEBUG) printf("%d:Input:%s",TEDfileCount,path);
 strcpy(TmpString,EDITFOLDER); strcat(TmpString,strchr(path,'/')); //check if substitution/correction PRG file is found in 'edits' folder
 if ( (InFile=fopen(TmpString,"rb")) == NULL ) { if ( (InFile=fopen(path,"rb")) == NULL ) return; }
 FileSize=0; while ( (DataIn=fgetc(InFile)) != EOF ) { FileData[FileSize]=DataIn; ++FileSize; }  fclose(InFile); if(DEBUG) printf(" (%d bytes) ",FileSize);
 strcpy(TargetPath,TARGETFOLDER); strcat(TargetPath,strchr(path,'/')); mkdir(dirname(TargetPath),0x777);
 strcpy(TargetPath,TARGETFOLDER); strcat(TargetPath,strchr(path,'/')); changeExt(TargetPath,TARGETEXTENSION);
 strcpy(TargetPath2,INFOTARGETFOLDER); strcat(TargetPath2,strchr(path,'/')); mkdir(dirname(TargetPath2),0x777);
 changeExt(path,INFOEXTENSION); strcpy(TargetPath2,INFOTARGETFOLDER); strcat(TargetPath2,strchr(path,'/'));

 for (i=0; i<sizeof(cRTED_TEDheader); ++i) ((char*)&TEDheader)[i]=0;
 changeExt(path,INFOEXTENSION); strcpy(InfoPath,INFOFOLDER); strcat(InfoPath,strchr(path,'/'));
 InfoFile = fopen(InfoPath,"r"); InfoFileSize=0;
 if(InfoFile!=NULL) { while ( (DataIn=fgetc(InfoFile)) != EOF ) { InfoText[InfoFileSize]=DataIn; ++InfoFileSize; } InfoText[InfoFileSize]='\0'; fclose(InfoFile); }
 else InfoText[0]='\0';
 strcpy(InfoPath,EDITFOLDER); strcat(InfoPath,strchr(path,'/'));
 EditFile = fopen(InfoPath,"r"); EditFileSize=0;
 if(EditFile!=NULL) { while ( (DataIn=fgetc(EditFile)) != EOF ) { EditText[EditFileSize]=DataIn; ++EditFileSize; } EditText[EditFileSize]='\0'; fclose(EditFile); }
 else EditText[0]='\0';
 LoadAddress = FileData[0] + (FileData[1]<<8); BASICoffset = BASICSTART_ADDRESS - LoadAddress; //Spektro music file-loadaddress = $1000 workaround

 //fill in TEDheader
 for (i=0; i<8; ++i) TEDheader.MagicString[i] = "TEDMUSIC"[i]; TEDheader.Version=0;
 for (Tmp=0, i=BASICoffset+6; /*i<32 &&*/ i<FileSize; ++i) if (FileData[i]==0x9E) break; //'SYS' BASIC-command found?
 if (/*i>=32 ||*/ i>=FileSize) Tmp = DEFAULT_INITADDRESS; //fallback trial if no SYS-command found
 else {
  for (++i; /*i<32 &&*/ i<FileSize; ++i) if ('0'<=FileData[i] && FileData[i]<='9') { Tmp = Tmp*10 + FileData[i]-'0'; }  else break;
 } if (Tmp==0x0000) Tmp = DEFAULT_INITADDRESS; //another fallback
 TEDheader.InitAddressL = Tmp&0xFF; TEDheader.InitAddressH = Tmp>>8;
 Tmp = DEFAULT_INITADDRESS; //Tmp = FileData[2+BASICoffset] + (FileData[3+BASICoffset]<<8) + 2; //can't count on FileData[2], e.g. Csabo's Paratroopers has something different there
 TEDheader.DataAddressL = Tmp&0xFF; TEDheader.DataAddressH = Tmp>>8;
 Tmp -= (BASICSTART_ADDRESS-2); TEDheader.DataOffsetL = Tmp&0xFF; TEDheader.DataOffsetH = Tmp>>8;
 TEDheader.PlayAddressL = TEDheader.PlayAddressH = 0x00; //get it from 'edits' folder for tunes that need/have it, especially multi-subtunes
 TEDheader.SubtuneAmount = 1; InfoPtr = findInfo("SubTunes",0);
 if (InfoPtr!=EmptyString) { sscanf(InfoPtr,"%d",&Tmp); TEDheader.SubtuneAmount=Tmp; }
 TEDheader.TimingSource = 0; TEDheader.Timer1valueL = 0x46; TEDheader.Timer1valueH = 0x45;
 TEDheader.SongFlags = 0; TEDheader.FileFlags = 0x01; //have a duration/playtime-block (and maybe keypress-block, especially for multi-subtune TEDs)
 Tmp = FileSize + sizeof(cRTED_TEDheader) - BASICoffset;
 if (TEDheader.FileFlags&1) { TEDheader.DurationOffsetL = Tmp&0xFF; TEDheader.DurationOffsetH = Tmp>>8; }
 for (i=0; i<TEDheader.SubtuneAmount; ++i) {
  sprintf(TmpString,"KeyPress%d",i+1); InfoPtr = findInfo(TmpString,0); if (InfoPtr!=EmptyString) TEDheader.FileFlags|=2;
 }
 //if InitAddress and PlayAddress is given (with single multispeed-call), convert RealTED to PlayTED
 InfoPtr = findInfo("InitAddress",0);
 if (InfoPtr!=EmptyString) {
  sscanf(InfoPtr,"%X",&Tmp); TEDheader.InitAddressL = Tmp&0xFF; TEDheader.InitAddressH = Tmp>>8;
  InfoPtr = findInfo("PlayAddress",0);
  if (InfoPtr!=EmptyString) {
   sscanf(InfoPtr,"%X",&Tmp); TEDheader.PlayAddressL = Tmp&0xFF; TEDheader.PlayAddressH = Tmp>>8;
   InfoPtr = findInfo("FrameSpeed",0);
   if (InfoPtr!=EmptyString) {
    sscanf(InfoPtr,"%dx",&Tmp);
    if (1<Tmp) {
     TEDheader.TimingSource = 2; if (Tmp>FRAMESPEED_MAX) Tmp=FRAMESPEED_MAX;
     TEDheader.Timer1valueL = FrameSpeedTimerLo[Tmp]; TEDheader.Timer1valueH = FrameSpeedTimerHi[Tmp];
 }}}}
 //fill in infotext-fields
 cutExt(path); if(compareFileExtension(path,".ted")) cutExt(path); fillInfoText( basename(path), "Title", TEDheader.Title);
 path = strchr(path,'/')+1; if(!strncmp(path,"musicians",9)) path = strchr(path,'/')+1;  *strrchr(path,'/') = '\0';
 fillInfoText(path, "Author", TEDheader.Author); fillInfoText("", "Date", TEDheader.ReleaseInfo);
 fillInfoText("", "Tool", TEDheader.Tool);

 //write out the essential music
 //don't modify if already a hiding TED with simple .PRG extension and the header fits, just update the header
 /*for (i=0; i<8; ++i) if (FileData[17+i]!="TEDMUSIC"[i]) break; //don't break but complete the missing field in place (if fits)?
 if (i==8) {
  OutFile = fopen(TargetPath,"wb"); if(OutFile==NULL) { printf("Can't create %s\n",TargetPath); return; }
  for (i=0; i<FileSize; ++i) fputc(FileData[i],OutFile); fclose(OutFile); printf("\n"); ++TEDfileCount; return;
 }*/
 Tmp = FileSize + BASICSTART_ADDRESS - BASICoffset;
 FileData[4+BASICoffset] = Tmp&0xFF; FileData[5+BASICoffset] = Tmp>>8; //rewrite BASIC SYS line-number for TED-format
 OutFile = fopen(TargetPath,"wb"); if (OutFile==NULL) { printf("Can't create %s\n",TargetPath); return; }
 if (BASICoffset==0) for (i=0; i<FileSize; ++i) fputc(FileData[i],OutFile);
 else { //fixate Spektro's $1000 load-address
  fputc(BASICSTART_ADDRESS&0xFF, OutFile); fputc(BASICSTART_ADDRESS>>8, OutFile);
  for(i=2+BASICoffset; i<FileSize; ++i) fputc(FileData[i],OutFile);
 }

 //write out TEDheader
 for (i=0; i<sizeof(cRTED_TEDheader); ++i) fputc( ((char*)&TEDheader)[17+i], OutFile);
 //write out duration-block
 if (TEDheader.FileFlags&1) {
  InfoPtr = findInfo("PlayTime",0); parseSavePlayTime();
  for (i=1; i<TEDheader.SubtuneAmount; ++i) {
   sprintf(TmpString,"PlayTime%d",i+1); InfoPtr = findInfo(TmpString,0); parseSavePlayTime();
  }
 }
 //write out keypress-block
 if (TEDheader.FileFlags&2) {
  InfoPtr = findInfo("KeyDelay",0); if (InfoPtr!=EmptyString) { sscanf(InfoPtr,"%d",&Tmp); Tmp&=0xFF; }  else Tmp = DEFAULT_KEYDELAY;
  fputc(Tmp,OutFile);
  for (i=0; i<TEDheader.SubtuneAmount; ++i) {
   sprintf(TmpString,"KeyPress%d",i+1); InfoPtr = findInfo(TmpString,0);
   if (InfoPtr!=EmptyString) {
    for (m=0;m<KEYMATRIX_SIZE;++m) { if (caseInsensitiveStringCompare(InfoPtr,KeyMatrixSymbols[m])) {fputc(m,OutFile); break;} }
    if (m==KEYMATRIX_SIZE) fputc(NOKEY,OutFile);
   }
   else fputc(NOKEY,OutFile);
 }}

 fclose(OutFile); if(DEBUG) printf("-> Output: %s\n",basename(TargetPath));
 ++TEDfileCount;

 //write out additional info (that has no field for in TED-format (or doesn't fit) )
 HVTCinfo[0]='\0'; saveInfoText("Title1"); saveInfoText("Notes"); saveInfoText("Covers");
 for (i=1; i<TEDheader.SubtuneAmount; ++i) {
  sprintf(TmpString,"Title%d",i+1); saveInfoText(TmpString);
  sprintf(TmpString,"Notes%d",i+1); saveInfoText(TmpString); sprintf(TmpString,"Covers%d",i+1); saveInfoText(TmpString);
 }
 if (HVTCinfo[0]!='\0') {
  OutFile2 = fopen(TargetPath2,"w");
  if (OutFile2==NULL) printf("Can't create %s\n",TargetPath2);
  else { fputs(HVTCinfo,OutFile2); fclose(OutFile2); }
 }
}


void createTEDfiles (char* basefolder, char* filename) {
 char Path[PATH_MAX_LENGTH]; static char DirPath[PATH_MAX_LENGTH]; DIR *Dir; struct dirent *Entry;
 Dir = opendir(basefolder); if (!Dir) return;
 while ( (Entry=readdir(Dir)) != NULL ) {
  if ( strcmp(Entry->d_name,".") && strcmp(Entry->d_name,"..") ) {
   strcpy(Path,basefolder); strcat(Path,"/"); strcat(Path,Entry->d_name);
   if ( compareFileExtension(Entry->d_name,".prg") && !compareFileExtension(Entry->d_name,".ted.prg")
        && (filename==NULL || !strcmp(Entry->d_name,filename)) ) convertToTEDfile(Path);
   else {
    strcpy(DirPath,TARGETFOLDER); strcat(DirPath,strchr(Path,'/')); mkdir(dirname(DirPath),0x777);
    strcpy(DirPath,INFOTARGETFOLDER); strcat(DirPath,strchr(Path,'/')); mkdir(dirname(DirPath),0x777);
    createTEDfiles(Path, filename);
   }
 }}
 closedir(Dir);
}


int main (char argc, char* argv[]) {

 if (argc==1 || !strcmp(argv[1],"-h")) { printf("Usage: HVTCted < html | playerfolder [individual-tunename.prg] >\n"); return 0; }
 else if (argc>=2) {
  if (!strcmp(argv[1],HTMLFOLDER)) {processHTMLinfo(); return 0;}
  createTEDfiles( argv[1], (argc>2)? argv[2]:NULL );
 }

 return 0;
}

